// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"fmt"
	"os"
	"path"
	"strings"

	"github.com/spf13/cobra"
	"github.com/spf13/viper"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/xbus/utils/license"
	"bitbucket.org/orus-io/xbus/version"
)

// Logger is the default logger for all commands.
// it is initialized early by initConfig(), so any code called
// by the Run function of a command is safe to use it.
var Logger xbus.Logger

var cfgFile string

// RootCmd represents the base command when called without any subcommands
var RootCmd = &cobra.Command{
	Use:   "xbusd",
	Short: "xbus main command",
	Long: `You certainly want to use one of our sub-commands

Xbus provides high level application messaging on top of NATS.`,
	Run: version.Cmd,
}

// Execute adds all child commands to the root command sets flags appropriately.
// This is called by main.main(). It only needs to happen once to the rootCmd.
func Execute() {
	if err := RootCmd.Execute(); err != nil {
		fmt.Println(err)
		os.Exit(-1)
	}
}

// preRunCheckCertDir verifies that certDir exists
func preRunCheckCertCertDir(cmd *cobra.Command, args []string) {
	// Ensure the cert directory exists.
	certDir := viper.GetString("cert-dir")
	if certDir == "" {
		if f := viper.ConfigFileUsed(); f != "" {
			certDir = path.Join(path.Dir(f), "certs", "xbusd")
		} else {
			certDir = "certs/xbusd"
		}
		viper.Set("cert-dir", certDir)
	}
	pathInfo, err := os.Stat(certDir)
	if err == nil && !pathInfo.IsDir() {
		err = fmt.Errorf("%q is not a directory", certDir)
	}
	if err != nil {
		Logger.Fatalf("Cannot access the cert directory at %q: %s", certDir, err)
	}
}

func init() {
	cobra.OnInitialize(initConfig)

	flags := RootCmd.PersistentFlags()
	version.InitFlags(flags)

	flags.StringVar(
		&cfgFile,
		"config", "",
		"config file (default is the first existing file in: $HOME/.config/xbus/xbusd.yaml, /etc/xbus/xbusd.yaml)",
	)

	flags.String(
		"cert-dir", "",
		"The directory where certificates and private keys of use to this program reside.",
	)

	flags.Float64(
		"tls-timeout", 4.0,
		"TLS handshake timeout",
	)

	flags.StringP(
		"database-dsn", "d", "",
		`PostgreSQL database connection string (ex: "sslmode=disable dbname=xbus")`,
	)

	flags.String(
		"nats-host", "localhost",
		"Host of the nats server",
	)

	flags.Int(
		"nats-port", 4222,
		"Port of the nats server",
	)

	flags.Int(
		"database-max-connections", 5,
		"Maximum allowed number of open database connections",
	)

	flags.Int(
		"storage-process-db-max-connections", 0,
		"Maximum number of db connections for the process storage (defaults to 30% of database.max-connections)",
	)

	flags.Bool(
		"debug", false,
		`Enable debugging mode, with lots of logging to the console`,
	)

	flags.Bool(
		"trace", false,
		`Enable even more logging than --debug.`,
	)

	viper.BindPFlag("cert-dir", flags.Lookup("cert-dir"))
	viper.BindPFlag("tls.timeout", flags.Lookup("tls-timeout"))
	viper.BindPFlag("nats.host", flags.Lookup("nats-host"))
	viper.SetDefault("nats.host", "localhost")
	viper.BindPFlag("nats.port", flags.Lookup("nats-port"))
	viper.SetDefault("nats.port", 4222)
	viper.BindPFlag("debug", flags.Lookup("debug"))
	viper.BindPFlag("trace", flags.Lookup("trace"))
	viper.BindPFlag("database.dsn", flags.Lookup("database-dsn"))
	viper.BindPFlag("database.max-connections", flags.Lookup("database-max-connections"))
	viper.BindPFlag(
		"storage.process.db.max-connections",
		flags.Lookup("storage-process-db-max-connections"))
}

func initLogger(status bool) {
	opts := xbus.LoggerOptions{
		Logtime: true,
		Debug:   viper.GetBool("debug"),
		Trace:   viper.GetBool("trace"),
	}
	Logger = xbus.GetLogger(&opts)
	if status {
		Logger.Debugf("Logger ready. Debug=%t, Trace=%t", opts.Debug, opts.Trace)
	}
}

// initConfig reads in config file and ENV variables if set.
func initConfig() {
	if cfgFile != "" { // enable ability to specify config file via flag
		viper.SetConfigFile(cfgFile)
	} else {
		viper.SetConfigName("xbusd") // name of config file (without extension)
	}

	envPrefix := "xbusd"
	homePath := "$HOME/.config/xbus/"
	etcPath := "/etc/xbus"
	viper.SetEnvPrefix(envPrefix) // our ENV vars must be prefixed by "XBUS_"

	viper.AutomaticEnv() // read in environment variables that match
	initLogger(false)
	Logger.Debugf("environment prefix is %s_", strings.ToUpper(envPrefix))

	Logger.Debugf("Adding to config path: %s", homePath)
	viper.AddConfigPath(homePath) // adding home directory as first search path
	Logger.Debugf("Adding to config path: %s", etcPath)
	viper.AddConfigPath(etcPath) // adding /etc/xbus as secondary search path

	// If a config file is found, read it in.
	err := viper.ReadInConfig()
	if err != nil {
		if err.Error() == "Unsupported Config Type \"\"" {
			Logger.Debugf("no configuration file found")
			Logger.Debugf("if you want a config file please add a xbusd.yaml into the config path")
		} else {
			Logger.Fatalf("error reading configuration file: %s\n", err)
		}
	} else {
		Logger.Debugf("Using config file: %s\n", viper.ConfigFileUsed())
	}
	initLogger(true)
	license.SetupLicenseWatcher(Logger)
}
