package control

import (
	"context"
	"crypto/x509"
	"encoding/pem"
	"fmt"
	"math/rand"
	"time"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/api/control"
	"bitbucket.org/orus-io/go-xbus/tools"
	"bitbucket.org/orus-io/xbus/storage"
)

type accountServer struct {
	*Service
}

var letters = []rune("0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ")

func randSeq(n int) string {
	b := make([]rune, n)
	for i := range b {
		b[i] = letters[rand.Intn(len(letters))]
	}
	return string(b)
}

// AccountAccept accepts a pending account request; a certificate is generated from the CSR.
func AccountAccept(
	accountStorage storage.AccountStorage, accountID api.UUID,
	serverCert *x509.Certificate, serverKey interface{},
	expire time.Time,
) (api.Account, error) {
	a, exists := accountStorage.Get(accountID)
	if !exists {
		return api.Account{}, fmt.Errorf("No such account: %s", accountID)
	}
	if a.Status != api.Account_PENDING {
		return api.Account{}, fmt.Errorf("Invalid initial status %s for 'Accept'", a.Status)
	}
	p, r := pem.Decode([]byte(a.CSR))
	if len(r) != 0 {
		return api.Account{}, fmt.Errorf("Invalid CSR pem encoding")
	}
	if p.Type != "CERTIFICATE REQUEST" {
		return api.Account{}, fmt.Errorf("Invalid CSR pem encoding")
	}
	cert, err := tools.CreateCertificateFromCSR(
		p.Bytes, serverCert, serverKey, expire,
	)
	if err != nil {
		return api.Account{}, err
	}
	a.Cert = string(pem.EncodeToMemory(&pem.Block{
		Type:  "CERTIFICATE",
		Bytes: cert,
	}))
	a.CSR = ""
	a.Status = api.Account_ENABLED
	return accountStorage.Update(a)
}

// AccountReject rejects a pending account request.
func AccountReject(
	accountStorage storage.AccountStorage, accountID api.UUID,
) (api.Account, error) {
	a, exists := accountStorage.Get(accountID)
	if !exists {
		return api.Account{}, fmt.Errorf("No such account: %s", accountID)
	}
	if a.Status != api.Account_PENDING {
		return api.Account{}, fmt.Errorf("Invalid initial status %q for 'Reject'", a.Status)
	}
	a.CSR = ""
	a.Status = api.Account_DISABLED
	return accountStorage.Update(a)
}

func accountDelete(storage storage.AccountStorage, accountID api.UUID) error {
	a, exists := storage.Get(accountID)
	if !exists {
		return xbus.ErrNotFound
	}
	storage.Delete(a)
	return nil
}

func (s accountServer) Delete(
	ctx context.Context, req control.AccountIDRequest,
) (err error) {
	err = accountDelete(s.o.AccountStorage, req.AccountID)
	if err != nil {
		if err != xbus.ErrNotFound {
			err = api.ClientProtoError("No such account")
		} else {
			err = api.ServerProtoError(err)
		}
		return
	}
	return
}

func (s accountServer) List(
	ctx context.Context, req control.AccountListRequest,
) (
	resp control.AccountListReply, err error,
) {
	resp.AccountList = s.o.AccountStorage.List()
	return
}

// HasPendingCSR returns true if the given account has a CSR pending
func HasPendingCSR(account api.Account) bool {
	return account.Status == api.Account_PENDING && account.CSR != ""
}

func (s accountServer) Update(
	ctx context.Context, req control.AccountUpdateRequest,
) (
	resp api.Account, err error,
) {
	if req.Account.ID.IsZero() {
		err = fmt.Errorf("Must pass an ID")
		return
	}
	a, exists := s.o.AccountStorage.Get(req.Account.ID)
	if !exists {
		err = fmt.Errorf("No such account: %s", req.Account.ID)
		return
	}
	s.Log.Noticef("Account: %#v", a)
	if HasPendingCSR(a) && req.Account.Status == api.Account_ENABLED {
		a, err = AccountAccept(
			s.o.AccountStorage, req.Account.ID, s.o.ServerCert, s.o.ServerKey, req.Expire)
	} else if HasPendingCSR(a) && req.Account.Status == api.Account_DISABLED {
		a, err = AccountReject(s.o.AccountStorage, req.Account.ID)
	} else {
		err = fmt.Errorf("Invalid change")
		return
	}
	if err != nil {
		return
	}
	resp = api.Account{
		ID:     a.ID,
		Name:   a.Name,
		Type:   a.Type,
		Status: a.Status,
		CSR:    a.CSR,
		Cert:   a.Cert,
	}
	return
}

func (s accountServer) RenewAPIKey(
	ctx context.Context, req control.AccountIDRequest,
) (
	account api.Account, err error,
) {
	account, ok := s.o.AccountStorage.Get(req.AccountID)
	if !ok {
		err = fmt.Errorf("No such account")
		return
	}

	account.APIKey = randSeq(32)
	account, err = s.o.AccountStorage.Update(account)
	if err != nil {
		err = api.ServerProtoError(err)
	}
	return
}

func (s accountServer) GetAPIKey(
	ctx context.Context, req control.AccountIDRequest,
) (
	account api.Account, err error,
) {
	account, ok := s.o.AccountStorage.Get(req.AccountID)
	if !ok {
		err = fmt.Errorf("No such account")
		return
	}
	return
}

func (s accountServer) SetAPIKey(
	ctx context.Context, req control.AccountUpdateRequest,
) (
	account api.Account, err error,
) {
	account, ok := s.o.AccountStorage.Get(req.Account.ID)
	if !ok {
		err = fmt.Errorf("No such account")
		return
	}

	account.APIKey = req.Account.APIKey
	account, err = s.o.AccountStorage.Update(account)
	if err != nil {
		err = api.ServerProtoError(err)
	}
	return
}

func (s accountServer) ResetCert(
	ctx context.Context, req control.AccountIDRequest,
) (
	err error,
) {
	account, ok := s.o.AccountStorage.Get(req.AccountID)
	if !ok {
		err = fmt.Errorf("No such account")
		return
	}

	if account.Status == api.Account_PENDING {
		return fmt.Errorf("The account is already in pending state. It must be accepted or rejected first")
	}

	if account.Status != api.Account_ENABLED && account.Status != api.Account_DISABLED {
		return fmt.Errorf("The account current state does now allow a cert reset. Expected ENABLED or DISABLED, got %s", account.Status)
	}

	account.Status = api.Account_PENDING
	account.CSR = ""

	account, err = s.o.AccountStorage.Update(account)

	return err
}
