// Copyright 2018 Orus.io. All rights reserved.

package storage

import (
	"testing"

	"github.com/stretchr/testify/assert"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
)

// JobStorageFactory is any factory of JobStorage
type JobStorageFactory func(bt testing.TB) (storage.JobStorage, func())

// JobStorageTests is a test suite for JobStorage implementations
func JobStorageTests(t *testing.T, factory JobStorageFactory) {

	t.Run("NewAndGet", func(t *testing.T) {
		js, clean := factory(t)
		defer clean()

		var job storage.Job
		job.JobID.ProcessID = api.MustUUIDFromString("3f5b2968-a6f2-11e8-a8ed-67a738169a27")
		job.JobID.Node.ActorID = api.MustUUIDFromString("483b8532-a6f2-11e8-b206-13d021cd09cd")
		job.JobID.Node.NodeID = "actor"
		job.JobState.Detached = false
		job.JobState.Status = api.ActorProcessingState_PENDING
		job.JobState.EnvelopeIDs = []api.UUID{
			api.MustUUIDFromString("204cc84c-f87d-11e8-99eb-4b1b847c458b"),
			api.MustUUIDFromString("2b48f2b6-f87d-11e8-b4f7-63d083cfc687"),
		}

		js.NewJob(job.JobID, job.JobState)

		state := js.GetState(job.JobID)
		assert.Equal(t, state, job.JobState)

		jobs := js.ListPendingJobByActor(job.Node.ActorID, nil, 1)
		assert.Equal(t, 1, len(jobs))
		if len(jobs) != 1 {
			t.Fail()
		}
		assert.Equal(t, jobs[0], job)

		jobs = js.ListJobByProcess(job.ProcessID)
		assert.Equal(t, 1, len(jobs))
		if len(jobs) != 1 {
			t.Fail()
		}
		assert.Equal(t, jobs[0], job)

		job.JobState.Status = api.ActorProcessingState_RUNNING
		js.SetState(job.JobID, job.JobState)

		jobs = js.ListRunningJobByEnvelopeID(
			api.MustUUIDFromString("204cc84c-f87d-11e8-99eb-4b1b847c458b"),
		)
		assert.Equal(t, 1, len(jobs))
		if len(jobs) != 1 {
			t.FailNow()
		}
		assert.Equal(t, jobs[0], job)

		jobs = js.ListRunningJobByEnvelopeID(
			api.MustUUIDFromString("3f5b2968-a6f2-11e8-a8ed-67a738169a27"),
		)
		assert.Equal(t, 0, len(jobs))

		job.JobState.Status = api.ActorProcessingState_SUCCESS
		js.SetState(job.JobID, job.JobState)

		jobs = js.ListPendingJobByActor(job.Node.ActorID, nil, 1)
		assert.Equal(t, 0, len(jobs))

		jobs = js.ListJobByProcess(job.ProcessID)
		assert.Equal(t, 0, len(jobs))

		state = js.GetState(job.JobID)
		assert.Equal(t, job.JobState, state)

	})

}
