package xbus

import (
	"bufio"
	"errors"
	"os"
	"path/filepath"
	"strings"
	"time"

	"github.com/nightlyone/lockfile"
)

// ErrInvalidKey is when an invalid key is passed to the API
var ErrInvalidKey = errors.New("Store keys cannot contain ':'")

// ErrCouldNotGetLock is when the store lock could not be acquired
var ErrCouldNotGetLock = errors.New("Could not get lock on the file")

// FilePersistentStore is a key-value store that:
// - is very safe (ie once Put is over the data will not get lost)
// - does not need to be fast for loading/reading (will be used only after crashes)
// - is multi-process safe
// TODO add a cleanup function that drop all deleted keys
type FilePersistentStore struct {
	path   string
	locker lockfile.Lockfile
}

// NewFilePersistentStore initializes a FilePersistentStore
func NewFilePersistentStore(path string) (*FilePersistentStore, error) {
	path, err := filepath.Abs(path)
	if err != nil {
		return nil, err
	}
	lockfname := path + ".lock"
	lock, err := lockfile.New(lockfname)
	if err != nil {
		return nil, err
	}
	return &FilePersistentStore{path, lock}, nil
}

func (s *FilePersistentStore) lock() error {
	// Get a lock
	for i := 0; i != 200; i++ {
		if i == 199 {
			return ErrCouldNotGetLock
		}
		err := s.locker.TryLock()
		if err == nil {
			return nil
		} else if _, ok := err.(interface {
			Temporary() bool
		}); ok {
			time.Sleep(5 * time.Millisecond)
		} else {
			return err
		}
	}
	panic("Unreachable")
}

// Put sets a value in the store
func (s *FilePersistentStore) Put(key, value string) error {
	if strings.Contains(key, ":") {
		return ErrInvalidKey
	}

	if err := s.lock(); err != nil {
		return err
	}
	defer s.locker.Unlock()

	// Open in append mode
	f, err := os.OpenFile(s.path, os.O_APPEND|os.O_WRONLY|os.O_CREATE, 0600)
	if err != nil {
		return err
	}
	defer f.Close()

	_, err = f.WriteString("put:" + key + ":" + value + "\n")
	if err != nil {
		return err
	}

	return nil
}

// Get retrieve a value from the store or "" if it does not exist
func (s *FilePersistentStore) Get(key string) (string, error) {
	if strings.Contains(key, ":") {
		return "", ErrInvalidKey
	}

	if err := s.lock(); err != nil {
		return "", err
	}
	defer s.locker.Unlock()

	_, err := os.Stat(s.path)
	if err != nil {
		if os.IsNotExist(err) {
			return "", nil
		}
		return "", err
	}
	// Open in readonly mode
	f, err := os.Open(s.path)
	if err != nil {
		return "", err
	}
	defer f.Close()

	currentvalue := ""

	// TODO read the file backward so we can stop reading as soon as an
	// entry matches
	scanner := bufio.NewScanner(f)
	for scanner.Scan() {
		line := scanner.Text()
		if strings.HasPrefix(line, "put:"+key+":") {
			currentvalue = line[5+len(key):]
		} else if strings.HasPrefix(line, "del:"+key) {
			currentvalue = ""
		}
	}

	if err := scanner.Err(); err != nil {
		return "", err
	}

	return currentvalue, nil
}

// Del removes a value from the store
func (s *FilePersistentStore) Del(key string) error {
	if strings.Contains(key, ":") {
		return ErrInvalidKey
	}

	if err := s.lock(); err != nil {
		return err
	}
	defer s.locker.Unlock()

	// Open in append mode
	f, err := os.OpenFile(s.path, os.O_APPEND|os.O_WRONLY|os.O_CREATE, 0600)
	if err != nil {
		return err
	}
	defer f.Close()

	_, err = f.WriteString("del:" + key + "\n")
	if err != nil {
		return err
	}

	return nil
}
