package xbus

import (
	"bytes"
	"crypto/ecdsa"
	"crypto/x509"
	"encoding/pem"
	"fmt"
	"io/ioutil"
	"os"
	"path"
	"path/filepath"
	"strconv"
	"strings"
	"time"

	"github.com/nats-io/go-nats"
	"github.com/spf13/pflag"
	"github.com/spf13/viper"
	"gopkg.in/yaml.v2"

	api "bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/tools"
)

const (
	confKeyAccountID   = "account-id"
	confKeyAccountName = "account-name"
	confKeyNatsURL     = "nats-url"
	confKeyNatsTimeout = "nats-timeout"
	confKeyForceStart  = "force-start"

	confKeyPersistentStore = "persistent-store"

	confKeyTLSCertfile     = "tls.certfile"
	confKeyTLSCertdir      = "tls.certdir"
	confKeyPrivateKey      = "tls.private-key"
	confKeyPrivateKeyFile  = "tls.private-key-file"
	confKeyCertificate     = "tls.certificate"
	confKeyCertificateFile = "tls.certificate-file"
	confKeyCSR             = "tls.csr"
	confKeyCSRFile         = "tls.csr-file"
	confKeyRootCA          = "tls.rootca"
	confKeyRootCAFile      = "tls.rootca-file"
	confKeyTLSTimeout      = "tls.timeout"

	confKeySentryDSN = "sentry.dsn"
)

func fileExists(path string) bool {
	if _, err := os.Stat(path); err != nil {
		if os.IsNotExist(err) {
			return false
		}
		panic(err)
	}
	return true
}

func readPEM(blockType string, pemOrPath string) []byte {
	var (
		bytes []byte
		err   error
	)
	if strings.Contains(pemOrPath, "-----BEGIN") {
		if !strings.Contains(pemOrPath, "-----BEGIN "+blockType+"-----") {
			panic(fmt.Sprintf(
				"Invalid pem block type. Expected %s, got:\n%s",
				blockType, pemOrPath))
		}
		bytes = []byte(pemOrPath)
	} else if fileExists(pemOrPath) {
		bytes, err = ioutil.ReadFile(pemOrPath)
		if err != nil {
			panic(err)
		}
	} else {
		panic("Not a file nor a PEM block: " + pemOrPath)
	}
	block, _ := pem.Decode(bytes)
	if block == nil {
		panic("Could not decode PEM data")
	}
	return block.Bytes
}

// ReadActor loads an actor options
func ReadActor(v *viper.Viper, key string) (ActorOptions, error) {
	name := v.GetString("name")
	if _, err := strconv.Atoi(key); err == nil {
		if name == "" {
			return ActorOptions{}, fmt.Errorf("Actor %s is missing a name", key)
		}
	} else {
		if name == "" {
			name = key
		} else if name != key {
			return ActorOptions{}, fmt.Errorf("Actor name inconsistent. Key is '%s', name is '%s'", key, name)
		}
	}
	var id api.UUID
	if s := v.GetString("id"); s != "" {
		id = api.MustUUIDFromString(s)
	}
	var aKind api.Actor_Kind
	if s := v.GetString("kind"); s != "" {
		if err := aKind.UnmarshalText([]byte(s)); err != nil {
			panic(err)
		}
	}
	var aStatus api.Actor_Status
	if s := v.GetString("status"); s != "" {
		if err := aStatus.UnmarshalText([]byte(s)); err != nil {
			panic(err)
		}
	}
	roles := v.GetStringSlice("roles")
	serviceType := v.GetString("service.type")
	serviceSettings := v.GetStringMap("service.settings")

	return ActorOptions{
		Actor: api.Actor{
			ID:     id,
			Name:   name,
			Kind:   aKind,
			Status: aStatus,
			Roles:  roles,
		},
		Service: ServiceOptions{
			Type:     serviceType,
			Settings: serviceSettings,
		},
	}, nil
}

// COptionsFromViper prepare a COptions from a Viper conf
// The TLS option changes are wrote back to the viper instance
func COptionsFromViper(
	v *viper.Viper,
	accountType api.Account_Type,
	varDir string,
	logger Logger) COptions {
	opts := COptions{
		Name:        v.GetString(confKeyAccountName),
		Type:        accountType,
		ForceStart:  v.GetBool(confKeyForceStart),
		NatsTimeout: v.GetDuration(confKeyNatsTimeout),
		ConnOptions: nats.Options{
			Url: v.GetString(confKeyNatsURL),
			Timeout: time.Duration(
				v.GetFloat64(confKeyTLSTimeout) * float64(time.Second)),
			AsyncErrorCB: func(c *nats.Conn, s *nats.Subscription, err error) {
				if s != nil {
					logger.Errorf("Error on sub %s: %s", s.Subject, err)
				} else {
					logger.Errorf("Error %s", err)
				}
			},
		},
	}

	if s := v.GetString(confKeyAccountID); s != "" {
		opts.ID = api.MustUUIDFromString(s)
	}

	var actorNames = make(map[string]bool)

	for _, key := range v.AllKeys() {
		if strings.HasPrefix(key, "actors.") {
			actor := strings.SplitN(key, ".", 3)[1]
			if _, ok := actorNames[actor]; ok {
				continue
			}
			logger.Noticef("Loading actor %s options", key)
			actorOpt, err := ReadActor(v.Sub("actors."+actor), actor)
			if err != nil {
				// TODO return a proper error
				panic(err)
			}
			opts.ActorList = append(opts.ActorList, actorOpt)
			actorNames[actor] = true
		}
	}

	if s := v.GetString(confKeyPrivateKey); s != "" {
		logger.Debugf("Loading pkey")
		key, err := x509.ParseECPrivateKey(readPEM("EC PRIVATE KEY", s))
		if err != nil {
			panic(err)
		}
		opts.TLS.PrivKey = key
	}

	if s := v.GetString(confKeyCSR); s != "" {
		logger.Debugf("Loading CSR")
		opts.TLS.CSR = readPEM("CERTIFICATE REQUEST", s)
	}

	if s := v.GetString(confKeyCertificate); s != "" {
		logger.Debugf("Loading certificate")
		opts.TLS.Certificate = readPEM("CERTIFICATE", s)
	}

	if s := v.GetString(confKeyRootCA); s != "" {
		logger.Debugf("Loading RootCA")

		opts.TLS.RootCAs = [][]byte{
			readPEM("CERTIFICATE", s),
		}
	}

	if s := v.GetString(confKeySentryDSN); s != "" {
		opts.SentryDSN = s
	}

	{
		var ckps string
		if ckps = v.GetString(confKeyPersistentStore); ckps == "" {
			ckps = varDir
		}
		store, err := NewFilePersistentStore(path.Join(ckps, "pstore.db"))
		if err != nil {
			panic(err)
		}
		opts.PersistentStore = store
	}

	return opts
}

// TLSKeyValueSaver saves the tls options as string key/value
type TLSKeyValueSaver struct {
	set func(key, value string) error
}

// NewTLSKeyValueSaver returns a TLSKeyValueSaver
func NewTLSKeyValueSaver(set func(key, value string) error) TLSKeyValueSaver {
	return TLSKeyValueSaver{set}
}

// SetPrivKey saves the private key
func (s TLSKeyValueSaver) SetPrivKey(key *ecdsa.PrivateKey) error {
	if key != nil {
		b, err := tools.PKeyToPEM(key)
		if err != nil {
			panic(err)
		}
		return s.set(confKeyPrivateKey, string(b))
	}
	return s.set(confKeyPrivateKey, "")
}

// SetCSR ...
func (s TLSKeyValueSaver) SetCSR(csr []byte) error {
	if csr != nil {
		b := pem.EncodeToMemory(&pem.Block{
			Type:  "CERTIFICATE REQUEST",
			Bytes: csr,
		})
		return s.set(confKeyCSR, string(b))
	}
	return s.set(confKeyCSR, "")
}

// SetCertificate ...
func (s TLSKeyValueSaver) SetCertificate(cert []byte) error {
	if cert != nil {
		b := pem.EncodeToMemory(&pem.Block{
			Type:  "CERTIFICATE",
			Bytes: cert,
		})
		return s.set(confKeyCertificate, string(b))
	}
	return s.set(confKeyCertificate, "")
}

// SetRootCAs ...
func (s TLSKeyValueSaver) SetRootCAs(cas [][]byte) error {
	if cas != nil {
		var buf bytes.Buffer
		for _, cert := range cas {
			pem.Encode(&buf, &pem.Block{
				Type:  "CERTIFICATE",
				Bytes: cert,
			})
		}
		return s.set(confKeyRootCA, buf.String())
	}
	return s.set(confKeyRootCA, "")
}

// TLSMapSaver saves a complete map of all the tls options each time an entry
// changes
func TLSMapSaver(initialOptions TLSOptions, save func(map[string]string) error) TLSOptionsSaver {
	var (
		ready = new(bool)
		m     = make(map[string]string)
	)

	saver := NewTLSKeyValueSaver(
		func(key, value string) error {
			m[key] = value
			if *ready {
				return save(m)
			}
			return nil
		},
	)

	saver.SetPrivKey(initialOptions.PrivKey)
	saver.SetCSR(initialOptions.CSR)
	saver.SetCertificate(initialOptions.Certificate)
	saver.SetRootCAs(initialOptions.RootCAs)

	*ready = true

	return saver
}

// TLSOptionsFileSaver returns a tlsSaver factory that writes the TLS options in
// a yaml file
func TLSOptionsFileSaver(
	certFile string, keySources map[string]string,
) (TLSOptionsSaver, error) {
	saver := NewTLSKeyValueSaver(
		func(key, value string) error {
			source, ok := keySources[key]
			if !ok {
				return fmt.Errorf("Unknown configuration entry: '%s'", key)
			}
			switch source {
			case "readonly":
				return fmt.Errorf(
					"Entry '%s' is read-only. "+
						"Try removing it from the main configuration file and put it in '%s'",
					key, certFile,
				)
			case "certfile":
				var values yaml.MapSlice
				b, err := ioutil.ReadFile(certFile)
				if err != nil && !os.IsNotExist(err) {
					return err
				}
				if err == nil {
					if err := yaml.Unmarshal(b, &values); err != nil {
						return err
					}
				}
				var found bool
				for i, item := range values {
					if item.Key.(string) == key {
						values[i].Value = value
						found = true
						break
					}
				}
				if !found {
					values = append(values, yaml.MapItem{Key: key, Value: value})
				}
				b, err = yaml.Marshal(values)
				if err != nil {
					return err
				}
				return ioutil.WriteFile(certFile, b, 0600)
			default:
				filePath := source[5:]
				dir := path.Dir(filePath)
				info, err := os.Stat(dir)
				if err != nil {
					if os.IsNotExist(err) {
						if err := os.MkdirAll(dir, 0700); err != nil {
							return err
						}
					} else {
						return err
					}
				} else if !info.Mode().IsDir() {
					return fmt.Errorf("certdir '%s' is not a directory", dir)
				}
				if strings.HasPrefix(source, "file:") {
					return ioutil.WriteFile(source[5:], []byte(value), 0600)
				}
				return fmt.Errorf("Unknown source type: %s", source)
			}
		},
	)
	return saver, nil
}

func loadTLSEntry(v *viper.Viper, key string, logger Logger) error {
	keyValue := v.GetString(key)
	fileValue := v.GetString(key + "-file")
	if keyValue != "" && fileValue != "" {
		return fmt.Errorf("you cannot define both '%s' and '%s'-file", key, key)
	}
	if fileValue != "" {
		b, err := ioutil.ReadFile(fileValue)
		switch {
		case err == nil:
			v.Set(key, string(b))
		case os.IsNotExist(err):
			logger.Noticef("%s: Ignoring inexistent file '%s'", key+"-file", fileValue)
		default:
			return fmt.Errorf("Could not read file '%s': %s", fileValue, err)
		}
	}
	return nil
}

func setDefaultTLSEntryFile(v *viper.Viper, key, filename string) {
	if v.GetString(key+"-file") == "" && v.GetString(key) == "" {
		v.Set(key+"-file", filename)
	}
}

func getTLSEntrySource(v *viper.Viper, key string) string {
	if v.GetString(key+"-file") != "" {
		return "file:" + v.GetString(key+"-file")
	}
	if v.GetString(key) != "" {
		return "readonly"
	}
	return "certfile"
}

// LoadTLSOptionFiles lookup the files storing the tls certs & keys and load
// them in the viper instance
func LoadTLSOptionFiles(v *viper.Viper, configPath string, logger Logger,
) (
	TLSOptionsSaver, error,
) {
	configDir := path.Dir(configPath)
	if !path.IsAbs(configDir) {
		p, err := filepath.Abs(configDir)
		if err != nil {
			return nil, err
		}
		configDir = p
	}
	accountName := v.GetString(confKeyAccountName)
	if accountName == "" {
		return nil, fmt.Errorf("Cannot do much without a '%s' defined", confKeyAccountName)
	}

	certdir := v.GetString(confKeyTLSCertdir)
	certfile := v.GetString(confKeyTLSCertfile)

	if certdir != "" && certfile != "" {
		return nil, fmt.Errorf(
			"you cannot define both '%s' and '%s'",
			confKeyTLSCertdir, confKeyTLSCertfile)
	}

	// If certdir is defined, use it to build filenames for entries without
	// value of file
	if certdir != "" {
		if !path.IsAbs(certdir) {
			certdir = path.Clean(
				path.Join(configDir, certdir),
			)
		}
		// TODO check if the directory exists (and is actually a directory)
		setDefaultTLSEntryFile(v, confKeyPrivateKey, path.Join(certdir, accountName+".key"))
		setDefaultTLSEntryFile(v, confKeyCertificate, path.Join(certdir, accountName+".crt"))
		setDefaultTLSEntryFile(v, confKeyCSR, path.Join(certdir, accountName+".csr"))
		setDefaultTLSEntryFile(v, confKeyRootCA, path.Join(certdir, "root.crt"))
	}

	// Now check if the options are already defined.
	// If defined as files, load them and make sure to keep the file path for
	// reuse when saving
	if err := loadTLSEntry(v, confKeyPrivateKey, logger); err != nil {
		return nil, err
	}
	if err := loadTLSEntry(v, confKeyCertificate, logger); err != nil {
		return nil, err
	}
	if err := loadTLSEntry(v, confKeyCSR, logger); err != nil {
		return nil, err
	}
	if err := loadTLSEntry(v, confKeyRootCA, logger); err != nil {
		return nil, err
	}

	var (
		keySources = map[string]string{
			confKeyPrivateKey:  getTLSEntrySource(v, confKeyPrivateKey),
			confKeyCertificate: getTLSEntrySource(v, confKeyCertificate),
			confKeyCSR:         getTLSEntrySource(v, confKeyCSR),
			confKeyRootCA:      getTLSEntrySource(v, confKeyRootCA),
		}
	)

	if certdir == "" && certfile == "" {
		// If certfile is not defined, auto-build a suitable path for it
		certfile = path.Join(configDir, accountName+".certs")
	}

	if certfile != "" {
		if !path.IsAbs(certfile) {
			certfile = path.Clean(
				path.Join(configDir, certfile),
			)
		}

		// Read the certfile if exists
		b, err := ioutil.ReadFile(certfile)
		if err != nil {
			// if the file does not exists, just ignore it
			// on other errors, stop everything
			if !os.IsNotExist(err) {
				return nil, fmt.Errorf("could not load '%s': %s", certfile, err)
			}
		} else {
			// TODO read the yaml file and make sure we don't override an existing
			// value
			var (
				values = make(map[string]string)
			)
			if err := yaml.Unmarshal(b, &values); err != nil {
				return nil, fmt.Errorf("Could not parse '%s': %s", certfile, err)
			}
			for _, key := range []string{confKeyPrivateKey, confKeyCertificate, confKeyCSR, confKeyRootCA} {
				if value, ok := values[key]; ok {
					if keySources[key] != "certfile" {
						return nil, fmt.Errorf("%s is redefined in '%s'", key, certfile)
					}
					v.Set(key, value)
				}

			}
		}
	}
	return TLSOptionsFileSaver(certfile, keySources)
}

// DefineAndBindPFlags populates a viper.Viper and pflags.FlagSet
// with all the options suitable for a xbus client
func DefineAndBindPFlags(v *viper.Viper, flags *pflag.FlagSet) {
	flags.StringP(confKeyNatsURL, "u", "nats://localhost:4222", "NATS url")
	v.BindPFlag(confKeyNatsURL, flags.Lookup(confKeyNatsURL))

	flags.Duration(confKeyNatsTimeout, 0, "NATS Timeout")
	v.BindPFlag(confKeyNatsTimeout, flags.Lookup(confKeyNatsTimeout))

	flags.String(confKeyAccountID, "", "Account ID")
	v.BindPFlag(confKeyAccountID, flags.Lookup(confKeyAccountID))

	flags.String(confKeyAccountName, "", "Account Name")
	v.BindPFlag(confKeyAccountName, flags.Lookup(confKeyAccountName))

	flags.String(confKeyPrivateKey, "", "Private key (a path or a PEM encoded key)")
	v.BindPFlag(confKeyPrivateKey, flags.Lookup(confKeyPrivateKey))

	flags.String(confKeyCertificate, "", "Certificate (a path or a PEM encoded cert)")
	v.BindPFlag(confKeyCertificate, flags.Lookup(confKeyCertificate))

	flags.String(confKeyCSR, "", "CSR (a path or a PEM encoded CSR)")
	v.BindPFlag(confKeyCSR, flags.Lookup(confKeyCSR))

	flags.Float64(confKeyTLSTimeout, 4.0, "TLS Timeout")
	v.BindPFlag(confKeyTLSTimeout, flags.Lookup(confKeyTLSTimeout))

	flags.Bool(
		confKeyForceStart, false,
		"Enable force start (stay up even if the server is not reachable on start)")
	v.BindPFlag(confKeyForceStart, flags.Lookup(confKeyForceStart))

	flags.String(confKeySentryDSN, "", "Sentry DSN")
	v.BindPFlag(confKeySentryDSN, flags.Lookup(confKeySentryDSN))
}

// COptionsFromFile load a client configuration from a file
func COptionsFromFile(configPath string, accountType api.Account_Type, logger Logger) (COptions, error) {
	in, err := os.Open(configPath)
	if err != nil {
		return COptions{}, err
	}

	xbusViper := viper.New()
	xbusFlags := &pflag.FlagSet{}

	DefineAndBindPFlags(xbusViper, xbusFlags)

	xbusViper.SetConfigType("yaml")
	if err := xbusViper.ReadConfig(in); err != nil {
		return COptions{}, err
	}

	configDir := path.Dir(configPath)

	if _, err := LoadTLSOptionFiles(xbusViper, configPath, logger); err != nil {
		return COptions{}, err
	}

	options := COptionsFromViper(
		xbusViper, accountType, configDir, logger,
	)

	return options, nil
}

// ClientFromFile loads a client from a configuration file
func ClientFromFile(path string, accountType api.Account_Type, logger Logger) (*Client, error) {
	options, err := COptionsFromFile(path, accountType, logger)

	if err != nil {
		return nil, err
	}

	client := NewClient(options, logger)
	return client, nil
}
